/* =========================================================================
	This library is part of IshiSoft application.
	------------------------------------------------------------------------
	author: Yuichiro Ishitani
			http://www001.upp.so-net.ne.jp/ishi/index.html
  ========================================================================== */
#include "stdafx.h"
#include "Str.h"
#include "commonhelper.h"
#include "MyConf.h"
#include "MyString.h"
#include "Usr.h"
#include "staticvariable.h"
#include "md5.h"
#include "stdio.h"

extern MyConf	g_conf;

static void hmac_md5( unsigned char* text, int text_len, unsigned char*  key, int key_len, char* digest );

Usr::Usr()
{

}

Usr::~Usr()
{

}

//-----------------------------------------------------------------------------
//	@detail
//		Get user home directory
//	@param
//		pszUsrName:			<user name> or <user name>@<localdomain>
//							mast have MAX_PATH length.
//		pszDir:				
//	@return
//		USR_LOCAL_DOMAIN
//		USR_OTHER_DOMAIN
//		USR_ERR_NAME
//		USR_ERR_BADPARAM	--> not error but not local user
//-----------------------------------------------------------------------------
int Usr::GetHomePath(const char *pszUsrName, char* pszDir)
{
	char		szUsrDir[MAX_PATH] = {0};
	char		szTmpUser[MAX_LOCALUSERNAME_LEN64+1] = {0};
	char*		pDomain = NULL;
	const char* pMe = g_conf.GetValueI(CONFIG_ME);
	const char* pszHome = g_conf.GetValueI(CONFIG_HOMEDIR);
	unsigned int	nUserNameLength;

	if( NULL == pszUsrName || NULL == pMe || NULL == pszHome )
		return USR_ERR_BADPARAM;

	STRNCPY( szTmpUser, pszUsrName, MAX_LOCALUSERNAME_LEN64 );

	pDomain = STRRCHR( szTmpUser, '@' );
	
	// chop atmark for get username.
	if( NULL != pDomain )
	{
		*pDomain++ = '\0';
	}

	// User name length must be MAX_USER_NAME(64) or less
	nUserNameLength = STRLEN( szTmpUser );
	if( nUserNameLength > MAX_LOCALUSERNAME_LEN64 || nUserNameLength == 0 )
	{
		return USR_ERR_NAME;
	}

	// Is this my domain?
	if( NULL != pDomain )
	{
		if( 0 != STRICMP( pDomain, pMe ) )
		{
			return USR_OTHER_DOMAIN;
		}
	}

	STRNCPY( szUsrDir, pszHome, MAX_CONFIG_DIR_PATH );
	STRNCAT( szUsrDir, szTmpUser, MAX_LOCALUSERNAME_LEN64 );

	if( NULL != pszDir )
	{
		STRNCPY( pszDir, szUsrDir, MAX_PATH );
	}
	return USR_LOCAL_DOMAIN;
}

//-----------------------------------------------------------------------------
//	@detail
//		Check the user is on this server.
//	@param
//		pszUsrName	user name
//	@return
//		USR_LOCAL_DOMAIN
//		USR_OTHER_DOMAIN
//		USR_ERR_NAME
//		USR_ERR_BADPARAM
//		USR_ERR_MEMORY
//-----------------------------------------------------------------------------
int Usr::IsMyUser( const char* pszUsrName )
{
	return GetHomePath(pszUsrName, NULL);
}

//-----------------------------------------------------------------------------
//	@detail
//		Resolve alias
//	@param
//		const char* pszDstDataLocalFile
//		MyHashEx *phsRcptLocal
//		MyHashFile *phsRcptRemote)
//	@return
//-----------------------------------------------------------------------------
int Usr::ResolveAlias(const char* pszDstDataLocalFile, MyHashEx *phsRcptLocal, MyHashFile *phsRcptRemote)
{
	MyString s;
	if( 0 == s.Set(pszDstDataLocalFile) )
	{
		return 0;
	}

	HANDLE hFile = CreateFile( s, GENERIC_READ, 0, NULL, OPEN_EXISTING, 0, NULL );

	if( hFile == INVALID_HANDLE_VALUE )
	{
		return 0;
	}

	char szMailAddress[MAX_MAILADDR_LEN256+2]={0};

	while( NULL != GetFileLine( hFile, szMailAddress, MAX_MAILADDR_LEN256 ) )
	{
		if( !GetMailAddressString( szMailAddress, MAX_MAILADDR_LEN256, szMailAddress, g_conf.GetValueI(CONFIG_ME) ) )
		{
			continue;
		}

		switch( IsMyUser( szMailAddress ) )
		{
		case USR_OTHER_DOMAIN:	//0
			phsRcptRemote->Set( szMailAddress, "" );
			break;

		case USR_LOCAL_DOMAIN:	//1
			phsRcptLocal->Set( szMailAddress, "" );
			break;

		default:
			break;
		}

	}

	CloseHandle( hFile );
	return 1;
}

//-----------------------------------------------------------------------------
//	@detail
//		Deliver mail file
//	@param
//		const char *srcFile
//		MyHashEx *phsRcptLocal
//		MyHashFile *phsRcptRemote
//	@return
//		0 error
//		1 SUCCESS
//-----------------------------------------------------------------------------
int Usr::DeliveryFile(const char *srcFile, MyHashEx *phsRcptLocal, MyHashFile *phsRcptRemote, MyLog* plog)
{
	const char* pszHome = g_conf.GetValueI(CONFIG_HOMEDIR);
	const char* pszQueueDir = g_conf.GetValueI(CONFIG_QUEUEDIR);
	MyString sSrcDataFile;
	
	if( 0 == sSrcDataFile.Set( srcFile ) )
	{
		return 0;
	}

	MyString sQueueFile;
	MyString sDstDataRemoteFile;
	MyString sDstDataLocalFile;

	if( NULL == srcFile || NULL == phsRcptLocal || NULL == phsRcptRemote || NULL == plog
		|| NULL == pszHome || NULL == pszQueueDir || 0 == STRLEN( sSrcDataFile ) )
		return 0;

	// safety check. is exist?
	DWORD dwFileAttr = GetFileAttributesMBCS( srcFile );

	char szDstDataFile[MAX_PATH] = {0};
	char szDstQueueFile[MAX_PATH] = {0};

	if( 0xFFFFFFFF == dwFileAttr )
		return 0;

	// get filename(remove filepath)
	const char* pFileName = srcFile;

	if( NULL == pszQueueDir )
		return 0;

	// move to last
	for( ; *pFileName != '\0'; pFileName++ );
	// move to end of backslash
	for( ; *pFileName != '\\' && pFileName != srcFile && pFileName != srcFile; pFileName-- );

	// ------------------------------------------------------------------------
	// resolve alias
	char szDbgMsg[512] = {0};
	char* pAtmark = NULL;
	char szLocalUserName[MAX_LOCALUSERNAME_LEN64+1] = {0};
	SEEKKEY skUsrName = phsRcptLocal->SeekToFirst();

	int nSafeCounter = 0;
	while( skUsrName )
	{
		// when loop counter over 10000, stop resolving alias.
		if( nSafeCounter++ > 10000 )
		{
			plog->Err( "alias loop found. some recipients dropped from recipient list." );
			break;
		}

		STRNCPY( szLocalUserName, skUsrName, MAX_LOCALUSERNAME_LEN64 );

		pAtmark = STRRCHR( szLocalUserName, '@' );
		
		if( NULL != pAtmark )
		{
			*pAtmark = '\0';
		}

		STRNCPY( szDstDataFile, pszHome, MAX_CONFIG_DIR_PATH );
		STRNCAT( szDstDataFile, szLocalUserName, MAX_LOCALUSERNAME_LEN64 );

		if( 0 == sDstDataLocalFile.Set( szDstDataFile ) )
		{
			skUsrName = phsRcptLocal->SeekNext( skUsrName );
			continue;
		}

		dwFileAttr = GetFileAttributes( sDstDataLocalFile );
		if( !(FILE_ATTRIBUTE_DIRECTORY & dwFileAttr) )
		{
			phsRcptLocal->Del( skUsrName );
			ResolveAlias(sDstDataLocalFile, phsRcptLocal, phsRcptRemote);
			skUsrName = phsRcptLocal->SeekToFirst();
			continue;
		}
		
		skUsrName = phsRcptLocal->SeekNext( skUsrName );
	}
	// ------------------------------------------------------------------------
	// start of remote delivery

	// check remote rcpt
	if( NULL != phsRcptRemote->SeekToFirst() )
	{
		STRNCPY( szDstQueueFile, pszQueueDir, MAX_CONFIG_DIR_PATH );
		STRCAT( szDstQueueFile, "index" );
		STRCAT( szDstQueueFile, pFileName );

		if( 0 == sQueueFile.Set( szDstQueueFile ) )
		{
			return 0;
		}

		if( 0 == phsRcptRemote->SaveFile( sQueueFile ) )
		{
			return 0;
		}

		// tmp file
		STRNCPY( szDstDataFile, pszQueueDir, MAX_CONFIG_DIR_PATH );
		STRCAT( szDstDataFile, pFileName );

		if( 0 == sDstDataRemoteFile.Set( szDstDataFile ) )
		{
			return 0;
		}

		if( 0 == STRLEN( sDstDataRemoteFile ) || 0 == CopyFile( sSrcDataFile, sDstDataRemoteFile, FALSE ) )
		{
			DeleteFile( sQueueFile );
			return 0;
		}
		plog->Dbg( "queing remote delivery file." );
	}
	// end of remote delivery
	// ------------------------------------------------------------------------
	// start of local delivery
	
	skUsrName = phsRcptLocal->SeekToFirst();

	while( skUsrName )
	{
		// make user file path
		STRNCPY( szLocalUserName, skUsrName, MAX_LOCALUSERNAME_LEN64 );

		pAtmark = STRRCHR( szLocalUserName, '@' );
		
		if( NULL != pAtmark )
		{
			*pAtmark = '\0';
		}

		STRNCPY( szDstDataFile, pszHome, MAX_CONFIG_DIR_PATH );
		STRNCAT( szDstDataFile, szLocalUserName, MAX_LOCALUSERNAME_LEN64 );

		STRCAT( szDstDataFile, pFileName );

		if( 0 == sDstDataLocalFile.Set( szDstDataFile ) )
		{
			goto LOCAL_DELIVERY_FAILED;
		}

		if( 0 == STRLEN( sDstDataLocalFile ) || 0 == CopyFile( sSrcDataFile, sDstDataLocalFile, FALSE ) )
		{
			goto LOCAL_DELIVERY_FAILED;
		}
		sprintf( szDbgMsg, "delivery to (%s).", skUsrName );
		plog->Info( szDbgMsg );
		skUsrName = phsRcptLocal->SeekNext( skUsrName );
	}
	return 1;
	// end of local delivery

LOCAL_DELIVERY_FAILED: // roll back

	// delete remote file.
	if( NULL != phsRcptRemote->SeekToFirst() )
	{
		DeleteFile( sQueueFile );
		DeleteFile( sDstDataRemoteFile );
	}

	// delete all local file.
	skUsrName = phsRcptLocal->SeekToFirst();
	while( skUsrName )
	{
		STRNCPY( szLocalUserName, skUsrName, MAX_LOCALUSERNAME_LEN64 );

		pAtmark = STRRCHR( szLocalUserName, '@' );
		
		if( NULL != pAtmark )
		{
			*pAtmark = '\0';
		}

		STRNCPY( szDstDataFile, pszHome, MAX_CONFIG_DIR_PATH );
		STRNCAT( szDstDataFile, szLocalUserName, MAX_LOCALUSERNAME_LEN64 );

		STRCAT( szDstDataFile, pFileName );

		sDstDataLocalFile.Set( szDstDataFile );
		DeleteFile( sDstDataLocalFile );
		sprintf( szDbgMsg, "delivery failed. delete message file for (%s).", skUsrName );
		plog->Info( szDbgMsg );
		skUsrName = phsRcptLocal->SeekNext( skUsrName );
	}

	return 0;
}


//-----------------------------------------------------------------------------
//	@detail
//		Check passwd
//	@param
//		const char *pszUserName
//		const char *pszPasswd
//	@return
//		0 error
//		1 OK
//-----------------------------------------------------------------------------
int Usr::AuthPlain(const char *pszUserName, const char *pszPasswd)
{
	MyHashFile cf;

	if( 0 == cf.LoadFile( g_conf.GetValueI(CONFIG_PASSWDFILE) ) )
		return 0;

	const char* pszRegisterdPasswd = cf.GetValueI( pszUserName );

	if( NULL == pszRegisterdPasswd || NULL == pszPasswd )
		return 0;

	if( 0 != STRCMP( pszRegisterdPasswd, pszPasswd ) )
	{
		return 0;
	}

	return 1;
}

//-----------------------------------------------------------------------------
//	@detail
//		Check passwd
//	@param
//		const char *pszUserName
//		const char *pszPasswd
//		const char* pszAPOPKey
//	@return
//		0 error
//		1 OK
//-----------------------------------------------------------------------------
int Usr::AuthAPOP(const char* pszUserName, const char* pClientDigest, const char* pszAPOPKey)
{
	if( NULL == pszUserName || NULL == pClientDigest || NULL == pszAPOPKey )
		return 0;

	MD5Context md5;
	unsigned char cDigest[16+2] = {0};

	// ---
	MyHashFile cf;

	if( 0 == cf.LoadFile( g_conf.GetValueI(CONFIG_PASSWDFILE) ) )
		return 0;

	const char* pszRegisterdPasswd = cf.GetValueI( pszUserName );

	if( NULL == pszRegisterdPasswd || NULL == pClientDigest )
		return 0;

	MD5Init(&md5);
	MD5Update(&md5, (const unsigned char*)pszAPOPKey, STRLEN( (const char*)pszAPOPKey ) );
	MD5Update(&md5, (const unsigned char*)pszRegisterdPasswd, STRLEN( (const char*)pszRegisterdPasswd ) );
	MD5Final( cDigest, &md5 );

	int ii;
	unsigned char cByte;
	for( ii = 0; ii < 16; ii++ )
	{
		cByte = 0;
		if( *pClientDigest > 'z' )		{return 0;}
		if( *pClientDigest >= 'a' )		{cByte = (*pClientDigest - 'a'+10) * 0x10;}
		else if( *pClientDigest >= 'A' ){cByte = (*pClientDigest - 'A'+10) * 0x10;}
		else if( *pClientDigest >= '0' ){cByte = (*pClientDigest - '0') * 0x10;}
		pClientDigest++;

		if( *pClientDigest > 'z' )		{return 0;}
		if( *pClientDigest >= 'a' )		{cByte += (*pClientDigest - 'a'+10);}
		else if( *pClientDigest >= 'A' ){cByte += (*pClientDigest - 'A'+10);}
		else if( *pClientDigest >= '0' ){cByte += (*pClientDigest - '0');}
		else							{return 0;}
		pClientDigest++;

		if( cDigest[ii] != cByte )
			return 0;
	}

	return 1;
}

//-----------------------------------------------------------------------------
//	@detail
//		Check passwd
//	@param
//		const char *pszPhrase
//		char *pszKey
//	@return
//		0 error
//		1 OK
//-----------------------------------------------------------------------------
int Usr::AuthCRAMMD5(char* pszPhrase, char* pszKey)
{
	if( NULL == pszPhrase || NULL == pszKey )
		return 0;

	MyHashFile cf;
	if( 0 == cf.LoadFile( g_conf.GetValueI(CONFIG_PASSWDFILE) ) )
		return 0;

	Base64Decode( pszPhrase );
	Base64Decode( pszKey );
	
	unsigned int nUserNameLen = 0;

	SEEKKEY sk = cf.SeekToFirst();
	while( NULL != sk )
	{
		nUserNameLen = STRLEN(sk);

		// phrase must have length that username + 16(digest)
		if( (nUserNameLen+16) > STRLEN(pszPhrase) )
			return 0;

		if( 0 == STRNICMP(sk, pszPhrase, nUserNameLen) )
			break;
		sk = cf.SeekNext(sk);
	}
	
	if( NULL == sk )
		return 0;

	const char* pszPasswd = cf.GetValueBySeekKey( sk );
	unsigned char cDigest[16+2] = {0};
	char szDigestStr[32+2] = {0};

	hmac_md5( (unsigned char*)pszKey, STRLEN( pszKey ), (unsigned char*)pszPasswd, STRLEN( pszPasswd ), (char*)cDigest );

	sprintf( szDigestStr, "%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x"
		,cDigest[0],cDigest[1],cDigest[2],cDigest[3],cDigest[4],cDigest[5],cDigest[6],cDigest[7]
		,cDigest[8],cDigest[9],cDigest[10],cDigest[11],cDigest[12],cDigest[13],cDigest[14],cDigest[15]);

	if( 0 != STRICMP( &pszPhrase[nUserNameLen+1], szDigestStr ) )
		return 0;

	return 1;
}

/*
RFC2104
HMAC: Keyed-Hashing for Message Authentication
text,                pointer to data stream
text_len,            length of data stream
key,                 pointer to authentication key
key_len,             length of authentication key
digest               caller digest to be filled in
*/
static void hmac_md5( unsigned char* text, int text_len, unsigned char*  key, int key_len, char* digest )
{
	MD5_CTX context;
	unsigned char tk[16];
	int i;
	/* if key is longer than 64 bytes reset it to key=MD5(key) */
	if (key_len > 64)
	{
		MD5_CTX      tctx;

		MD5Init(&tctx);
		MD5Update(&tctx, key, key_len);
		MD5Final(tk, &tctx);

		key = tk;
		key_len = 16;
	}

	/*
	 * the HMAC_MD5 transform looks like:
	 *
	 * MD5(K XOR opad, MD5(K XOR ipad, text))
	 *
	 * where K is an n byte key
	 * ipad is the byte 0x36 repeated 64 times

	 * opad is the byte 0x5c repeated 64 times
	 * and text is the data being protected
	 */

	/* start out by storing key in pads */
	unsigned char k_ipad[65] = {0};
	unsigned char k_opad[65] = {0};
	strcpy( (char*)k_ipad, (const char*)key );
	strcpy( (char*)k_opad, (const char*)key );
	//bcopy( key, k_ipad, key_len);
	//bcopy( key, k_opad, key_len);


	/* XOR key with ipad and opad values */
	for (i=0; i<64; i++) {
		k_ipad[i] ^= 0x36;
		k_opad[i] ^= 0x5c;
	}
	/*
	 * perform inner MD5
	 */
	MD5Init(&context);							/* init context for 1st pass */
	MD5Update(&context, k_ipad, 64);			/* start with inner pad */
	MD5Update(&context, text, text_len);		/* then text of datagram */
	MD5Final((unsigned char*)digest, &context);	/* finish up 1st pass */
	/*
	 * perform outer MD5
	 */
	MD5Init(&context);                   /* init context for 2nd
										  * pass */
	MD5Update(&context, k_opad, 64);     /* start with outer pad */
	MD5Update(&context, (unsigned char*)digest, 16);     /* then results of 1st
										  * hash */
	MD5Final((unsigned char*)digest, &context);          /* finish up 2nd pass */
}
